// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_TAG_CONFIG_HXX
#define MPD_TAG_CONFIG_HXX

struct ConfigData;

void
TagLoadConfig(const ConfigData &config);

#endif
