// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_INPUT_SMBCLIENT_H
#define MPD_INPUT_SMBCLIENT_H

extern const struct InputPlugin input_plugin_smbclient;

#endif
