// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_FFMPEG_DOMAIN_HXX
#define MPD_FFMPEG_DOMAIN_HXX

class Domain;

extern const Domain ffmpeg_domain;

#endif
