// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_DECODER_SNDFILE_HXX
#define MPD_DECODER_SNDFILE_HXX

extern const struct DecoderPlugin sndfile_decoder_plugin;

#endif
