// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#pragma once

class TagMask;

/**
 * These are the tags that are allowed to have stickers.
 */
extern const TagMask sticker_allowed_tags;
