// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_ARCHIVE_ZZIP_HXX
#define MPD_ARCHIVE_ZZIP_HXX

struct ArchivePlugin;

extern const ArchivePlugin zzip_archive_plugin;

#endif
