// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_OUTPUT_API_HXX
#define MPD_OUTPUT_API_HXX

// IWYU pragma: begin_exports

#include "OutputPlugin.hxx"
#include "Interface.hxx"
#include "pcm/AudioFormat.hxx"
#include "tag/Tag.hxx"
#include "config/Block.hxx"

// IWYU pragma: end_exports

#endif
