// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#include "Init.hxx"

void
SmbclientInit()
{
	/* this is currently a no-op, but one day, we might want to
	   call smbc_thread_posix() here */
}
