// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_DECODER_ADPLUG_H
#define MPD_DECODER_ADPLUG_H

extern const struct DecoderPlugin adplug_decoder_plugin;

#endif
