// SPDX-License-Identifier: GPL-2.0-or-later
// Copyright The Music Player Daemon Project

#ifndef MPD_NEIGHBOR_UPNP_HXX
#define MPD_NEIGHBOR_UPNP_HXX

struct NeighborPlugin;

extern const NeighborPlugin upnp_neighbor_plugin;

#endif
